#!/usr/bin/env python3
"""
Main module for Where is Ben Nevis (it's in Scotland).
"""

#
# Version number
#
from ._nevis_version import (   # noqa
    __version__,
    __version_tuple__,
)

#
# Check Python version
#
import sys
if sys.hexversion < 0x03060000:
    raise RuntimeError('nevis requires Python 3.6 or newer.')
del sys    # Don't expose as part of API

#
# Installed project root
#
import os, inspect  # noqa
try:
    frame = inspect.currentframe()
    _DIR_MODULE = os.path.abspath(os.path.dirname(inspect.getfile(frame)))
    _DIR_MODULE_DATA = os.path.join(_DIR_MODULE, '_bin')
finally:
    # Always manually delete frame
    # https://docs.python.org/2/library/inspect.html#the-interpreter-stack
    del frame
del os, inspect    # Don't expose as part of API

#
# Terrain data
#
import os   # noqa

_ENV_DATA = 'NEVIS_PATH'
_DIR_DATA = os.environ.get(_ENV_DATA)
if _DIR_DATA is None:
    _DIR_DATA = os.path.join('~', 'nevis-data')
_DIR_DATA = os.path.abspath(os.path.expanduser(_DIR_DATA))
del os


#
# Create public API
#
from ._bng import (    # noqa
    ben,
    Coords,
    dimensions,
    fen,
    Hill,
    macdui,
    pub,
    schiehallion,
    squares,
)
from ._os_terrain_50 import (    # noqa
    DataNotFoundError,
    download_os_terrain_50,
    gb,
    inland_below_sea_level,
    is_sea,
    spacing,
)
from ._interpolation import (   # noqa
    linear_interpolant,
    spline,
)
from ._plot import (    # noqa
    plot,
    plot_line,
    save_plot,
)
from ._util import (    # noqa
    generate_kml,
    howdy,
    print_result,
    Timer,
)


#
# Test method
#
def write_test_figure(path='gb-small.png'):
    """
    Creates and writes a test figure.
    """
    gb()
    labels = {'Ben Nevis': ben(), 'Holme Fen': fen()}
    fig, ax, heights, g = plot(labels=labels, zoom=1 / 32)
    save_plot(path, fig, verbose=True)




#!/usr/bin/env python3
"""
Module providing methods to work with BNG (OS36GB) data, specifically the 700
by 1300km grid covering GB.
"""
import csv
import os
import random
import urllib
import zipfile

import bnglonlat
import numpy as np
import scipy.spatial

import nevis

# Get accurate longitude/lattitude, or use fallback
try:
    import convertbng.util

    def lonlat(x, y):
        a, b = convertbng.util.convert_lonlat([x], [y])
        if not (np.isnan(a) or np.isnan(b)):
            return a[0], b[0]
        return bnglonlat.bnglonlat(x, y)

except ImportError:
    lonlat = bnglonlat.bnglonlat


# Full size of the grid (in meters): 700km by 1300km
# Origin is a bottom left (in SV square)
width, height = 700000, 1300000


# Hill names
hill_zip = os.path.join(nevis._DIR_MODULE_DATA, 'hills.zip')
hill_file = 'hills.csv'

# Grid letters (bottom to top, left to right)
GL = [
    ['V', 'W', 'X', 'Y', 'Z'],
    ['Q', 'R', 'S', 'T', 'U'],
    ['L', 'M', 'N', 'O', 'P'],
    ['F', 'G', 'H', 'J', 'K'],
    ['A', 'B', 'C', 'D', 'E'],
]


def dimensions():
    """ Returns the dimensions of the grid (width, height) in meters. """
    return width, height


def squares():
    """
    Returns a list of tuples ``(name, x, y)`` with two-letter 100km grid square
    names and lower-left corners, covering the area of the data.
    """
    squares = []

    d = 100000
    i0, i1 = 1, 0   # Start on 2nd row (QRSTU)
    for y in range(0, height, d):
        j0, j1 = 2, 0   # Start on 2nd letter (S,N)
        for x in range(0, width, d):
            squares.append((GL[i0][j0] + GL[i1][j1], x, y))
            j1 += 1
            if j1 == 5:
                j1 = 0
                j0 += 1
        i1 += 1
        if i1 == 5:
            i1 = 0
            i0 += 1
    return squares


class Coords(object):
    """
    Coordinates, either normalised (so that 0, 0 is the bottom left of the grid
    and 1, 1 is the top right) or as OS36GB grid points in meters.

    Examples::

        p = nevis.Coords(gridx=123456, gridy=123456)
        print(p.grid)
        print(p.normalised)

        p = nevis.Coords(normx=0.5, normy=0.2)
        print(p.grid)
        print(p.normalised)

        b = nevis.ben()
        print(b.square)
        print(b.geograph)

    """

    def __init__(self, gridx=None, gridy=None, normx=None, normy=None):
        self._gridx = self._gridy = None
        self._normx = self._normy = None

        if gridx is not None and gridy is not None:
            if normx is None and normy is None:
                self._gridx = int(gridx)
                self._gridy = int(gridy)
                self._normx = self._gridx / width
                self._normy = self._gridy / height
        if normx is not None and normy is not None:
            if gridx is None and gridy is None:
                self._normx = float(normx)
                self._normy = float(normy)
                self._gridx = int(self._normx * width)
                self._gridy = int(self._normy * height)
        if self._gridx is None:
            raise ValueError(
                'Either (gridx and gridy) or (normx and normy) must be'
                'specified (and not both).')

        self._latlong = None
        self._square = None
        self._square3 = None
        self._square4 = None
        self._square5 = None

    def _find_square(self, n):

        # Get letters and remaining x y
        if self._square is None:
            # Lower-left of "V" (the bottom-left major letter) is 1000km to the
            # left of the origins of the grid (SV) and 500km below it.
            x, y = self._gridx + 1000000, self._gridy + 500000

            # Get first letter
            a, b = int(y // 500000), int(x // 500000)
            try:
                if a < 0 or b < 0:
                    raise KeyError
                name = GL[a][b]

                # Get second letter
                x, y = x % 500000, y % 500000
                a, b = int(y // 100000), int(x // 100000)
                name += GL[a][b]

                # Get numbers
                x, y = x % 100000, y % 100000
                x, y = int(x), int(y)

                # Make string
                self._square = name, f'{x:0>5}', f'{y:0>5}'
            except KeyError:
                self._square = False

        # Make string
        if self._square is False:
            return 'Off the grid'

        name, x, y = self._square
        return name + x[:n] + y[:n]

    @staticmethod
    def from_square(square):
        """
        Creates coordinates corresponding to the lower-left corner of a grid
        square indicated by letters and numbers, e.g. ``NN166712`` or
        ``NN 166 712``.
        """
        return Coords.from_square_with_size(square)[0]

    @staticmethod
    def from_square_with_size(square):
        """
        Like :meth:`from_square` but returns a tuple ``(coords, size)`` where
        ``size`` is the length of the square's sides.
        """
        code = square.strip().upper()
        if len(code) < 1:
            raise ValueError(
                'Invalid BNG grid reference: must be at least 1 letter')

        def find(letter):
            """
            Find lower-left coordinates. Big squares are 500km, small 100km.
            """
            for i, row in enumerate(GL):
                if letter >= row[0] and letter != 'I':
                    return i, row.index(letter)
            raise ValueError(
                f'Invalid BNG grid letter "{letter}" in code "{square}".')

        # Origin of letter system is V, BNG starts at S
        x, y = -1000000, -500000

        # Process letters
        i, j = find(code[0])
        x, y = x + 500000 * j, y + 500000 * i
        if len(code) == 1:
            return Coords(x, y), 500000

        i, j = find(code[1])
        x, y = x + 100000 * j, y + 100000 * i
        if len(code) == 2:
            return Coords(x, y), 100000

        # Check numbers
        numbers = code[2:].split(maxsplit=1)
        if len(numbers) == 1:
            numbers = numbers[0]
            n = len(numbers)
            if n % 2 == 1:
                raise ValueError(
                    f'Invalid BNG grid code: {square} numbers must have same'
                    ' number of digits.')
            n = n // 2
            numbers = (numbers[:n], numbers[n:])
        else:
            n = max(len(numbers[0]), len(numbers[1]))

        try:
            a, b = int(numbers[0]), int(numbers[1])
        except ValueError:
            raise ValueError(
                f'Invalid BNG grid code: {square} numbers must be integers.')
        if a < 0 or b < 0:
            raise ValueError(
                f'Invalid BNG grid code: {square} numbers must be positive.')

        # Calculate square size
        m = 10**(5 - n)

        # Final grid coordinates
        x, y = x + a * m, y + b * m

        # Pretty much always, we'll want to round
        if m >= 1:
            x, y = int(x), int(y)

        return Coords(x, y), m

    @property
    def grid(self):
        return np.array([self._gridx, self._gridy])

    @property
    def normalised(self):
        return np.array([self._normx, self._normy])

    @property
    def square3(self):
        if self._square3 is None:
            self._square3 = self._find_square(3)
        return self._square3

    @property
    def square4(self):
        if self._square4 is None:
            self._square4 = self._find_square(4)
        return self._square4

    @property
    def square5(self):
        if self._square5 is None:
            self._square5 = self._find_square(5)
        return self._square5

    @property
    def latlong(self):
        if self._latlong is None:
            lon, lat = lonlat(self._gridx, self._gridy)
            self._latlong = lat, lon
        return self._latlong

    @property
    def geograph(self):
        return f'http://www.geograph.org.uk/gridref/{self.square5}'

    @property
    def google(self):
        lat, long = self.latlong
        lat, long = round(lat, 6), round(long, 6)
        return (
            'https://www.google.com/maps/@?api=1&map_action=map'
            f'&center={lat},{long}&zoom=15&basemap=terrain')

    @property
    def osmaps(self):
        lat, long = self.latlong
        lat, long = round(lat, 6), round(long, 6)
        return (
            f'https://explore.osmaps.com/en/pin?lat={lat}&lon={long}&zoom=17')

    @property
    def opentopomap(self):
        lat, long = self.latlong
        lat, long = round(lat, 6), round(long, 6)
        return f'https://opentopomap.org/#marker=15/{lat}/{long}'

    def __str__(self):
        return f'Coords({int(round(self._gridx))}, {int(round(self._gridy))})'


class Hill(object):
    """
    Known hill tops.

    Examples::

        print(Hill.by_name('Ben Nevis'))
        print(Hill.by_rank(1))
        print(Hill.by_rank(2))
        print(Hill.by_rank(3))

        h, d = Hill.nearest(Coords(gridx=216600, gridy=771300))
        print(h)

    """
    _hills = []
    _names = {}
    _ids = {}
    _tree = None

    def __init__(self, x, y, rank, height, hill_id, name):
        if Hill._tree is not None:
            raise ValueError('Cannot add hills after tree construction.')
        self._x = int(x)
        self._y = int(y)
        self._rank = int(rank)
        self._height = float(height)
        self._id = int(hill_id)
        self._name = name.strip()
        self._photo = None
        Hill._hills.append(self)
        Hill._names[self._name.lower()] = self
        Hill._ids[self._id] = self

    @staticmethod
    def _load():
        """ Loads hills into memory. """

        # Unzip
        with zipfile.ZipFile(hill_zip, 'r') as f:
            with f.open(hill_file, 'r') as g:
                lines = g.read().decode('utf-8')
        lines = lines.splitlines()
        rows = iter(csv.reader(lines, delimiter=',', quotechar='"'))

        # Parse header
        fields = ['x', 'y', 'rank', 'meters', 'id', 'name']
        header = next(rows)
        try:
            indices = [header.index(field) for field in fields]
        except ValueError as e:
            raise ValueError(f'Unable to read hill-file header: {e}')

        # Parse data
        coords = []
        for row in rows:
            Hill(*[row[i] for i in indices])
            coords.append([row[indices[0]], row[indices[1]]])

        # Construct tree
        Hill._tree = scipy.spatial.KDTree(np.array(coords, dtype=np.float32))

    @staticmethod
    def by_id(hill_id):
        """
        Return a hill with the given ``hill_id`` (as used on e.g. hill
        bagging.)
        """
        if not Hill._hills:
            Hill._load()
        return Hill._ids[hill_id]

    @staticmethod
    def by_name(name):
        """ Return a hill with the given ``name``. """
        if not Hill._hills:
            Hill._load()
        return Hill._names[name.lower()]

    @staticmethod
    def by_rank(rank):
        """
        Return the hill with the given ``rank`` (rank 1 is heighest, then 2,
        etc.).
        """
        if not Hill._hills:
            Hill._load()
        if rank < 1 or rank > len(Hill._hills):
            raise ValueError(f'Rank outside of range: {rank}.')
        hill = Hill._hills[rank - 1]
        assert hill.rank == rank, 'Hills not ordered by rank'
        return hill

    @staticmethod
    def nearest(coords):
        """
        Returns a tuple (hill, distance) with the hill nearest to the given
        points.
        """
        if not Hill._hills:
            Hill._load()
        d, h = Hill._tree.query([coords._gridx, coords._gridy])
        return Hill._hills[h], d

    @property
    def coords(self):
        return Coords(gridx=self._x, gridy=self._y)

    @property
    def height(self):
        return self._height

    @property
    def hill_id(self):
        return self._id

    @property
    def name(self):
        return self._name

    @property
    def rank(self):
        return self._rank

    @property
    def ranked(self):
        s = str(self._rank)
        if s[-1] in '123':
            if not (self._rank > 10 and s[-2] == '1'):
                return s + 'st' if s[-1] == '1' else s + 'd'
        return s + 'th'

    @property
    def summit(self):
        return f'http://hillsummits.org.uk/htm_summit/{self._id}.htm'

    @property
    def portrait(self):
        return f'http://hillsummits.org.uk/htm_portrait/{self._id}.htm'

    def photo(self):
        """
        Attempts to return a URL with a photo. Returns an empty string if none
        is found.
        """
        def status(url):
            request = urllib.request.Request(url, method='HEAD')
            try:
                with urllib.request.urlopen(request) as f:
                    status = f.status
            except urllib.error.HTTPError:
                status = 404
            return status

        if self._photo is None:
            if status(self.summit) != 404:
                self._photo = self.summit
            elif status(self.portrait) != 404:
                self._photo = self.portrait
            else:
                self._photo = ''
        return self._photo

    def __str__(self):
        return f'{self._name} ({self.height}m)'


def ben():
    """ Returns the coordinates of Ben Nevis. """
    return Coords.ben


def fen():
    """ Returns the coordinates of Holme Fen """
    return Coords.fen


def pub(name=None):
    """ Returns coordinates where a mathematician may be found. """
    if name:
        return Coords.pub[name]
    return random.choice(list(Coords.pub.values()))


Coords.ben = Coords(gridx=216666, gridy=771288)
Coords.fen = Coords(gridx=520483, gridy=289083)
Coords.pub = {
    'Bear': Coords(gridx=451473, gridy=206135),
    'Canal house': Coords(gridx=457307, gridy=339326),
    'MacSorleys': Coords(gridx=258809, gridy=665079),
    'Sheffield tap': Coords(gridx=435847, gridy=387030),
    'Ship Inn': Coords(gridx=293945, gridy=72712),
    'Laurel Inn': Coords(gridx=495227, gridy=505037),
}


def schiehallion():
    """
    Returns a tuple ``(hill, boundaries)`` containing the :class:`Hill` object
    for Schiehallion and boundaries ``(x_lower, x_upper, y_lower, y_upper)``
    defining its neighbourhood.

    This can be used as an easier test problem for local methods.
    """
    hill = nevis.Hill.by_name('Schiehallion')
    x, y = hill.coords.grid
    b = 3e3
    boundaries = [x - b * .7, x + b * 1.1, y - b * .55, y + b * .5]
    return hill, boundaries


def macdui():
    """
    Returns a tuple ``(hill, boundaries)`` containing the :class:`Hill` object
    for Ben Macdui and boundaries ``(x_lower, x_upper, y_lower, y_upper)``
    defining its neighbourhood, Cairngorms Mountains.

    This can be used as an easier test problem for local methods.
    """
    hill = nevis.Hill.by_name('Ben Macdui')
    x, y = hill.coords.grid
    b = 8e3
    boundaries = [x - b * 1.75, x + b * 2.5, y - b * 1.4, y + b * 1.9]
    return hill, boundaries



#!/usr/bin/env python3
"""
Provides plotting methods.
"""
# The zip contains several grid squares, using the lettering described here:
# https://en.wikipedia.org/wiki/Ordnance_Survey_National_Grid
#
# Inspired by
# https://scipython.com/blog/processing-uk-ordnance-survey-terrain-data/
#
import warnings

import matplotlib.colors
import matplotlib.figure
import numpy as np

import nevis


def plot(boundaries=None, labels=None, trajectory=None, points=None,
         scale_bar=True, big_grid=False, small_grid=False,
         zoom=1 / 27, headless=False, verbose=False):
    """
    Creates a plot of the 2D elevation data in ``heights``.

    By default, this method uses ``pyplot`` to create the figure, which can
    have the side-effect of instantiating a back-end for maptlotlib.
    Alternatively, the plot can be created in "headless" mode, by setting
    ``headless=True``.

    Arguments:

    ``boundaries``
        An optional tuple ``(xmin, xmax, ymin, ymax)`` defining the boundaries
        (in meters) of the plotted region.
    ``labels``
        An optional dictionary mapping string labels to points (tuples in
        meters or Coords objects) that will be plotted on the map (if within
        the boundaries).
    ``trajectory``
        An optional array of shape ``(n_points, 2)`` indicating the trajectory
        followed to get to Ben Nevis (points specified in meters).
    ``points``
        An optional array of shape ``(n_points, 2)`` indicating points on the
        map (points specified in meters).
    ``scale_bar``
        Set to ``False`` to disable the scale bar.
    ``big_grid``
        Show the 2-letter grid squares (100km by 100km).
    ``small_grid``
        Show the 2-letter 2-number grid squares (10km by 10km)
    ``zoom``
        Adjust plot size by interpolating (``zoom > 1``) or downsampling (
        ``zoom < 1``).

        To downsample by e.g. a factor 10, set ``zoom = 1 / 10``. To
        interpolate and show e.g. 3x3 pixels per data point, set ``zoom = 3``.
        Interpolation is performed using matplotlib's "bilinear interpolation".

        The default value is ``1 / 27``, which creates a reasonably sized plot
        for the full GB data set.
    ``headless``
        Set to ``True`` to create the figure without using pyplot.
    ``verbose``
        Set to ``True`` to print updates to ``stdout`` while plotting.

    Returns a tuple ``(fig, ax, heights, g)`` where ``fig`` is the created
    figure, ``ax`` is the axes the image is plotted on, and ``heights`` is the
    (possibly downsampled) numpy array. The final entry ``g`` is a function
    that converts coordinates in meters to coordinates on the map axes.
    """
    # Current array shape
    heights = nevis.gb()
    ny, nx = heights.shape

    # Get extreme points (before any downsampling!)
    vmin = np.min(heights)
    vmax = np.max(heights)
    if verbose:
        print(f'Lowest point: {vmin}')
        print(f'Highest point: {vmax}')

    # Calculate downsampling using zoom
    if zoom > 1:
        downsampling = 1
    else:
        downsampling = int(round(1 / zoom))
        zoom = 1

    # Downsample (27 gives me a map that fits on my screen at 100% zoom).
    if downsampling > 1:
        if verbose:
            print(f'Downsampling with factor {downsampling}')
        heights = heights[::downsampling, ::downsampling]
        ny, nx = heights.shape

    # Select region to plot, and create meters2indices method
    d_org = d_new = np.array(nevis.dimensions())    # In meters
    offset = np.array([0, 0])                       # In meters
    if boundaries is not None:
        xlo, xhi, ylo, yhi = [float(x) for x in boundaries]

        # Select appropriate part of array
        xlo = max(0, int(xlo / d_org[0] * nx))
        ylo = max(0, int(ylo / d_org[1] * ny))
        xhi = min(nx, int(np.ceil(xhi / d_org[0] * nx)))
        yhi = min(ny, int(np.ceil(yhi / d_org[1] * ny)))

        heights = heights[ylo:yhi, xlo:xhi]

        # Adjust array size
        ny, nx = heights.shape

        # Set new dimensions and origin (bottom left)
        r = nevis.spacing() * downsampling
        d_new = np.array([nx * r, ny * r])
        offset = np.array([xlo * r, ylo * r])

    def meters2indices(x, y):
        """ Convert meters to array indices (which equal image coordinates) """
        x = (x - offset[0]) / d_new[0] * nx
        y = (y - offset[1]) / d_new[1] * ny
        try:
            x, y = int(x), int(y)
        except TypeError:
            x, y = x.astype(int), y.astype(int)
        return x, y

    # Plot
    if verbose:
        print('Plotting...')

    # Create colormap
    # f = absolute height, g = relative to vmax (and zero)
    f = lambda x: (x - vmin) / (vmax - vmin)
    # g = lambda x: f(x * vmax)
    cmap = matplotlib.colors.LinearSegmentedColormap.from_list(
        'soundofmusic', [
            (0, '#4872d3'),             # Deep sea blue
            (f(-0.1), '#68b2e3'),       # Shallow sea blue
            (f(0.0), '#0f561e'),        # Dark green
            (f(10), '#1a8b33'),         # Nicer green
            (f(100), '#11aa15'),        # Glorious green
            (f(300), '#e8e374'),        # Yellow at ~1000ft
            (f(610), '#8a4121'),        # Brownish at ~2000ft
            (f(915), '#999999'),        # Grey at ~3000ft
            (1, 'white'),
        ], N=1024)
    #import matplotlib.cm
    #cmap = matplotlib.cm.get_cmap('inferno')

    # Work out figure dimensions
    # Note: Matplotlib defaults to 100 dots per inch and 72 points per inch for
    # font sizes and line widths. This means that increasing the dpi leads to
    # more pixels per inch, but also to much bigger letters and thicker lines,
    # as it assumes the physical size should stay the same when printed!
    dpi = 100
    fw = nx * zoom / dpi
    fh = ny * zoom / dpi
    if verbose:
        print(f'Figure dimensions: {fw}" by {fh}" at {dpi} dpi')
        print(f'Should result in {int(fw * dpi)} by {int(fh * dpi)} pixels.')

    # Create figure
    if headless:
        fig = matplotlib.figure.Figure(figsize=(fw, fh), dpi=dpi)
    else:
        import matplotlib.pyplot as plt
        fig = plt.figure(figsize=(fw, fh), dpi=dpi)
    fig.subplots_adjust(0, 0, 1, 1)

    # Add axes
    ax = fig.add_subplot(1, 1, 1)
    ax.set_axis_off()
    ax.imshow(
        heights,
        origin='lower',
        cmap=cmap,
        vmin=vmin,
        vmax=vmax,
        interpolation='bilinear' if zoom > 1 else 'none',
    )
    ax.set_xlim(0, nx)
    ax.set_ylim(0, ny)

    # Add grid
    if small_grid:
        for sq, x0, y0 in nevis.squares():
            for j in range(10):
                for i in range(10):
                    x, y = x0 + j * 10000, y0 + i * 10000
                    if y == 0 and x > 0:
                        q, r = meters2indices(x, y)
                        if q > 2 and q < nx - 2:
                            ax.axvline(q, color='w', lw=0.5)
                    elif x == 0 and y > 0:
                        q, r = meters2indices(x, y)
                        if r > 2 and r < ny - 2:
                            ax.axhline(r, color='w', lw=0.5)
                    q, r = meters2indices(x + 5000, y + 5000)
                    if q > 10 and q < nx - 10 and r > 10 and r < ny - 10:
                        ax.text(q, r, sq + str(j) + str(i), color='w',
                                ha='center', va='center', fontsize=10)
    elif big_grid:
        for sq, x, y in nevis.squares():
            if y == 0 and x > 0:
                q, r = meters2indices(x, y)
                if q > 0 and q < nx:
                    ax.axvline(q, color='w', lw=0.5)
            elif x == 0 and y > 0:
                q, r = meters2indices(x, y)
                if r > 0 and r < ny:
                    ax.axhline(r, color='w', lw=0.5)
            q, r = meters2indices(x + 50000, y + 50000)
            if q > 20 and q < nx - 20 and r > 10 and r < ny - 10:
                ax.text(q, r, sq, color='w',
                        ha='center', va='center', fontsize=14)

    # Add scale bar
    if scale_bar:
        # Guess a good size
        x = d_new[0] / 5
        if x > 250e3:
            x = int(round(x / 250e3) * 250e3)
        elif x > 90e3:
            x = int(round(x / 100e3) * 100e3)
        elif x > 9e3:
            x = int(round(x / 10e3) * 10e3)
        elif x > 4.5e3:
            x = int(round(x / 5e3) * 5e3)
        elif x > 1e3:
            x = int(round(x / 1e3) * 1e3)
        else:
            x = int(round(x / 100) * 100)
        t = f'{x}m' if x < 1000 else f'{x // 1000}km'
        x = x / d_new[0] * nx
        y = 0.05 * ny
        dy = 0.015 * ny
        x0, x1 = 0.5 * x, 1.5 * x
        ax.plot([x0, x1], [y, y], 'white', lw=1)
        ax.plot([x0, x0], [y - dy, y + dy], 'white', lw=1)
        ax.plot([x1, x1], [y - dy, y + dy], 'white', lw=1)
        ax.text(0.5 * (x0 + x1), y + 0.5 * dy, t, color='white',
                horizontalalignment='center')

    # Show requested points
    if points is not None:
        points = np.asarray(points)
        alpha, ms, mw = (0.3, 4, 1) if len(points) > 20 else (1, 12, 2)
        x, y = meters2indices(points[:, 0], points[:, 1])
        ax.plot(x, y, 'x', color='#0000ff',
                markeredgewidth=mw, markersize=ms, alpha=alpha)

    # Show trajectory
    if trajectory is not None:
        trajectory = np.asarray(trajectory)
        x, y = meters2indices(trajectory[:, 0], trajectory[:, 1])
        ax.plot(
            x, y, 'o-', color='#000000',
            lw=0.5, markeredgewidth=0.5, markersize=4)

    # Add labelled points
    if labels:
        n_plotted = 0
        kwargs = {'fillstyle': 'none', 'markersize': 12}
        for label, p in labels.items():
            if isinstance(p, nevis.Coords):
                p = p.grid
            x, y = meters2indices(*p)
            if x > 0 and x < nx and y > 0 and y < ny:
                n_plotted += 1
                ax.plot(x, y, 'wo', markeredgewidth=3, **kwargs)
                ax.plot(x, y, 'o', markeredgewidth=2, label=label, **kwargs)

        if n_plotted:
            ax.legend(
                loc='upper left',
                framealpha=1,
                handlelength=1.5,
                handletextpad=0.9,
            )

    return fig, ax, heights, meters2indices


def plot_line(f, point_1, point_2, label_1='Point 1', label_2='Point 2',
              padding=0.25, evaluations=400, figsize=(8, 5), headless=False,
              verbose=False):
    """
    Draws a line between two points and evaluates a function along it.

    Arguments:

    ``f``
        A function ``f(x, y) -> z``, or a sequence of multiple such functions.
    ``point_1``
        The first point as a set of Coords or a numpy array in meters.
    ``point_2``
        The second point.
    ``label_1``, ``label_2``
        Optional labels for the points.
    ``padding``
        The amount of padding shown to the left and right of the points, as a
        fraction of the total line length (e.g. ``padding=0.25`` extends the
        line on either side by 25% of the distance between the two points).
    ``evaluations``
        The number of evaluations of ``f`` to plot.
    ``figsize``
        The default figure size
    ``headless``
        Set to ``True`` to create the figure without using pyplot, i.e. to save
        with ``figsave`` instead of calling ``plt.show()``.
    ``verbose``
        Set to ``True`` to print updates to ``stdout`` while plotting.

    Returns a tuple ``(fig, ax, p1, p2)`` where ``fig`` is the generated
    figure, ``ax`` is the axes object within that figure, and ``p1`` and ``p2``
    are the extremities of the line (points 1 and 2 plus padding) (as Coords).
    """
    # Points as vectors
    if isinstance(point_1, nevis.Coords):
        point_1 = point_1.grid
    if isinstance(point_2, nevis.Coords):
        point_2 = point_2.grid

    # Direction vector
    r = point_2 - point_1
    d = np.sqrt(r[0]**2 + r[1]**2)

    # Points to evaluate
    s = np.linspace(-padding, 1 + padding, evaluations)
    p = [point_1 + sj * r for sj in s]

    # Functions
    if callable(f):
        fs = [f]
    else:
        fs = f
        for i, f in enumerate(fs):
            if not callable(f):
                raise ValueError(
                    'f must be a callable or a sequence of callables: found'
                    f' non-callable at index {i}.')

    # Evaluations-es
    ys = [[f(*x) for x in p] for f in fs]

    # Create figure
    if headless:
        fig = matplotlib.figure.Figure(figsize=figsize)
    else:
        import matplotlib.pyplot as plt
        fig = plt.figure(figsize=figsize)

    # Plot
    fig.subplots_adjust(0.1, 0.1, 0.99, 0.99)
    ax = fig.add_subplot(1, 1, 1)
    ax.set_xlabel('Distance (m)')
    ax.set_ylabel('Altitude - according to our interpolation (m)')
    ax.spines['right'].set_visible(False)
    ax.spines['top'].set_visible(False)
    for k, y in enumerate(ys):
        ax.plot(s * d, y, label=f'Function {1 + k}')
    ax.axvline(0, ls='-', color='r', label=label_1)
    ax.axvline(d, ls='--', color='k', label=label_2)
    ax.axvspan(0, d, color='#fffede', lw=0, zorder=-1)
    ax.legend()

    return fig, ax, nevis.Coords(*p[0]), nevis.Coords(*p[-1])


def save_plot(path, fig, heights=None, verbose=False):
    """
    Stores the given figure using ``fig.savefig(path)``.

    If ``heights`` is given and ``PIL`` (pillow) is installed it will also
    check that the image dimensions (in pixels) equal the size of ``heights``.
    """
    if verbose:
        print(f'Writing figure to {path}')
    fig.savefig(path)

    # Try importing PIL to check image size
    if heights is None:
        return
    try:
        import PIL
    except ImportError:
        return

    # Suppress "DecompressionBomb" warning
    PIL.Image.MAX_IMAGE_PIXELS = None

    # Open image, get file size
    if verbose:
        print('Checking size of generated image')
    with PIL.Image.open(path) as im:
        ix, iy = im.size

    if (iy, ix) == heights.shape:
        if verbose:
            print('Image size OK')
    else:
        warnings.warn(
            f'Unexpected image size: width {ix}, height {iy}, expecting'
            f' {heights.shape}.')




#!/usr/bin/env python3
"""
Module providing interpolation methods over the OS 50 data set.
"""
import os
import pickle

import numpy as np
import scipy.interpolate

import nevis


class linear_interpolant(object):
    """
    Returns a linear interpolation and optionally its gradient over the full
    GB data set.

    When ``grad`` is set to ``False`` (by default), the returned function
    takes two arguments ``x`` and ``y`` (both in metres) and returns an
    interpolated height ``z`` (in meters).

    When ``grad`` is set to ``True``, the returned function takes two
    arguments ``x`` and ``y`` (both in metres) and returns a tuple ``(z, g)``,
    where ``z`` is an interpolated height (in meters) and ``g`` is a tuple
    ``(dz/dx, dz/dy)``.

    The height for each grid point ``(i, j)`` is assumed to be in the center of
    the square from ``(i, j)`` to ``(i + 1, j + 1)``.

    Example::

        f = linear_interpolation()
        print(f(1000, 500))

        f_grad = linear_interpolation(grad=True)
        z, (gx, gy) = f_grad(1000, 500)
        print(f"Height: {z:.2f} m, gradient: ({gx:.2f} m/m, {gy:.2f} m/m)")

    """
    # Note: This is technically a class, but used as a function here so
    # following the underscore naming convention.

    def __init__(self, grad=False):
        self._heights = nevis.gb()
        self._resolution = nevis.spacing()
        self._grad = grad

    def __call__(self, x, y):
        ny, nx = self._heights.shape
        x, y = x / self._resolution - 0.5, y / self._resolution - 0.5

        # Find nearest grid points
        # x1 (left), x2 (right), y1 (bottom), y2 (top).

        # Nearest grid points
        # When outside the grid, we use the _two_ points nearest the edge
        # (resulting in an extrapolation)
        x1 = np.minimum(nx - 2, np.maximum(0, int(x)))
        y1 = np.minimum(ny - 2, np.maximum(0, int(y)))
        x2 = x1 + 1
        y2 = y1 + 1

        # Heights at nearest grid points (subscripts are x_y)
        h11 = self._heights[y1, x1]
        h12 = self._heights[y2, x1]
        h21 = self._heights[y1, x2]
        h22 = self._heights[y2, x2]

        # We omit the 1 / (x2 - x1) and 1 / (y2 - y1) terms as these are always
        # 1 in the normal (interpolating) case.
        # When extrapolating, we get e.g. x1 == x2, so that
        # (x2 - x) == (x1 - x)

        # X-direction interpolation on both y values
        f1 = np.where(h11 == h21, h11, (x2 - x) * h11 + (x - x1) * h21)
        f2 = np.where(h12 == h22, h12, (x2 - x) * h12 + (x - x1) * h22)

        # Final result
        f = float(np.where(f1 == f2, f1, (y2 - y) * f1 + (y - y1) * f2))
        if self._grad:
            # Gradient of the interpolant
            g = (h21 - h11) / self._resolution, (h12 - h11) / self._resolution
            return f, g
        else:
            return f


def spline(verbose=False):
    """
    Returns a spline interpolation over the full GB data set.

    The returned function takes two arguments ``x`` and ``y`` (both in metres)
    and returns an interpolated height ``z`` (in meters).

    Example::

        f = spline()
        print(f(1000, 500))

    Notice: Calling this method will result in the creation and storage of a
    very large cache file in the nevis data directory.
    """
    heights = nevis.gb()

    # Load pickled spline
    s = None
    cached = os.path.join(nevis._DIR_DATA, 'spline')
    if os.path.isfile(cached):
        if verbose:
            print('Loading cached spline...')
        try:
            with open(cached, 'rb') as f:
                s = pickle.load(f)
        except Exception:
            if verbose:
                print('Loading failed.')

    # Create new spline
    if s is None:
        if verbose:
            print('Reticulating splines...')
        width, height = nevis.dimensions()
        ny, nx = heights.shape
        c = 25  # Correction: Coords at lower-left, height is center of square
        t = nevis.Timer()
        s = scipy.interpolate.RectBivariateSpline(
            np.linspace(0, height, ny, endpoint=False) + c,
            np.linspace(0, width, nx, endpoint=False) + c,
            heights,
        )
        if verbose:
            print(f'Completed in {t.format()}')

        # Cache to disk
        if verbose:
            print('Caching spline to disk...')
            t = nevis.Timer()
        with open(cached, 'wb') as f:
            pickle.dump(s, f)
        if verbose:
            print(f'Completed in {t.format()}')

    return lambda x, y: s(y, x)[0][0]



#!/usr/bin/env python3
"""
Hidden module that provides methods to download, unpack, and process the OS
Terrain 50 data set.
"""
# The zip contains several grid squares, using the lettering described here:
# https://en.wikipedia.org/wiki/Ordnance_Survey_National_Grid
#
# Inspired by
# https://scipython.com/blog/processing-uk-ordnance-survey-terrain-data/
#
import io
import os
import sys
import urllib.request
import zipfile

# Patch the zipfile module to support the PKZIP proprietary DEFLATE64 format
# This is used by two files in the May 2022 release of OS Terrain 50
# (data/sd/sd67_OST50GRID_20220506.zip and data/sd/sd68_OST50GRID_20220506.zip)
import zipfile_deflate64  # noqa

import numpy as np

import nevis


# Base name of the big zip file to read (e.g. to read 'terr50_gagg_gb.zip' we
# put 'terr50_gagg_gb').
terrain_file = 'terr50_gagg_gb'
terrain_file_zip = os.path.join(nevis._DIR_DATA, terrain_file + '.zip')
terrain_file_npy = os.path.join(nevis._DIR_DATA, terrain_file + '.npy')
not_sea_file = 'not_sea'
not_sea_file_npy = os.path.join(nevis._DIR_DATA, not_sea_file + '.npy')

# URL to download it from
url = ('https://api.os.uk/downloads/v1/products/Terrain50/downloads?'
       'area=GB&format=ASCII+Grid+and+GML+%28Grid%29&redirect')

# Resolution: distance between neighbours in the grid
resolution = 50

# .asc file encoding
ENC = 'utf-8'

# Cached heights and not-sea-mask
_heights = None
_not_sea = None             # y * width + x
_not_sea_unpacked = None    # array of y, x indices


class DataNotFoundError(RuntimeError):
    """ Raised when the OS Terrain 50 data is not found. """


def spacing():
    """ Returns the spacing between any two grid points. """
    return resolution


def gb(downsampling=None):
    """
    Returns an array of elevation data for Great Britain, with height in meters
    and grid points spaced 50m apart.

    Array indices are (height, width), also known as (northings, eastings).

    A downsampled version can be returned for testing purposes, by setting
    ``downsampling`` to any integer greater than one.
    """
    global _heights, _not_sea

    # Load data (or retrieve from cache)
    if _heights is None or _not_sea is None:
        # Check files exists
        if not os.path.isfile(terrain_file_npy):
            raise DataNotFoundError(
                f'OS Terrain 50 data not found in {nevis._DIR_DATA}.'
                ' Please use nevis.download_os_terrain_50() to download and'
                ' process this data set.'
            )
        # Don't load the heights data yet, may have to be rebuild it anyway!
        if not os.path.isfile(not_sea_file_npy):
            raise DataNotFoundError(
                'List of below sea-level inland points not found in'
                f' {nevis._DIR_DATA}. Please call'
                ' nevis.download_os_terrain_50() to create this file.'
            )

        # Load both files
        _heights = np.load(terrain_file_npy)
        _heights.setflags(write=False)
        _not_sea = np.load(not_sea_file_npy)
        _not_sea.setflags(write=False)

    # Return downsampled version for testing (but keep full version in cache)
    if downsampling is not None:
        downsampling = int(downsampling)
        if downsampling > 1:
            return _heights[::downsampling, ::downsampling]

    return _heights


def inland_below_sea_level_indices():
    """
    Returns a list of indices in the heights data that are below sea-level, but
    which nevis's preprocessing did not consider to be "sea".
    """
    global _not_sea_unpacked

    if _not_sea_unpacked is None:
        if _not_sea is None:
            gb()
        x = _not_sea % _heights.shape[1]
        y = _not_sea // _heights.shape[1]
        _not_sea_unpacked = np.vstack((y, x)).T

    return _not_sea_unpacked


def inland_below_sea_level():
    """
    Returns a list of points (x, y) in meters that are below sea-level, but
    which nevis's preprocessing did not consider to be "sea".
    """
    ij = inland_below_sea_level_indices() * resolution
    return np.vstack((ij[:, 1], ij[:, 0])).T


def is_sea(coords):
    """
    Returns ``True`` if and only if the given :class:`nevis.Coords` were
    classified as "sea" by nevis's pre-processing.
    """
    if _heights is None:
        gb()
    x, y = coords.grid // nevis.spacing()
    try:
        if _heights[y, x] >= 0:
            return False
    except IndexError:  # Anything off-map counts as sea
        return True
    return (y * _heights.shape[1] + x) not in _not_sea


def download_os_terrain_50(force=False):
    """
    Downloads, unpacks and processes the OS Terrain 50 data set.

    If a previously downloaded zip file is found then the downloading step is
    skipped, unless ``force`` is set to ``True``.

    If a previously unpacked and processed file is found, this method does
    nothing unless ``force`` is set to ``True``.
    """
    global _heights

    # Already done and not forcing? Then return
    have_terrain = os.path.isfile(terrain_file_npy)
    have_not_sea = os.path.isfile(not_sea_file_npy)
    if (not force) and have_terrain and have_not_sea:
        print('Downloaded, unpacked, and processed file already found:'
              ' Skipping.')
        return

    # Check directory exists, show message saying it will be created and how to
    # change its location.
    make_dirs = not os.path.isdir(nevis._DIR_DATA)
    if make_dirs:
        msg = (
            f'This method will create a directory: {nevis._DIR_DATA}\n'
            'This will be used to store the OS Terrain 50 data set in both a'
            ' compressed (160mb) and uncompressed (1.5gb) form, as well as'
            ' additional files that can reach several gb in size.\n'
            'These files will NOT be deleted automatically when nevis is'
            ' uninstalled.\n'
        )
        if nevis._ENV_DATA in os.environ:
            msg += 'This path was'
        else:
            msg += 'An alternative directory can be'
        msg += f' specified with the environment variable: {nevis._ENV_DATA}'
        print(msg)

    # Download zip file
    if force or not os.path.isfile(terrain_file_zip):
        print('The OS Terrain 50 database will be downloaded from')
        print(f'  {url}')
        print()
        yesno = input('Continue? (y/n) ')
        if yesno.lower() not in ['y', 'yes']:
            print('Halted.')
            return

        if make_dirs:
            os.makedirs(nevis._DIR_DATA)
        download_gagg(url, terrain_file_zip)

    # Unpack / process
    if force or not have_terrain or not have_not_sea:
        # Temporary sea-mask level
        s = -100

        # Create empty array
        width, height = nevis.dimensions()
        nx, ny = width // resolution, height // resolution
        heights = np.empty((ny, nx), dtype=np.float32)
        heights.fill(np.nan)

        # Fill it up
        extract(terrain_file, heights, resolution)

        # Replace missing values by far-below-sea level
        heights[np.isnan(heights)] = s

        # Fix odd squares
        fix_sea_levels_in_odd_squares(heights)

        # Add a few fake damns
        save_cambridgeshire(heights)

        # Create a "sea mask" in the loaded data, indicating sea with -100
        set_sea_level(heights, s)

        # Store below-zero points that are not in the sea, so that we can
        # later detect if something has been labelled sea or not.
        not_sea = inland_below_sea_level_points(heights, s)
        print(f'Saving to {not_sea_file_npy}...')
        np.save(not_sea_file_npy, not_sea)

        # Make the points labelled as sea slope towards the land, to make it
        # more findable.
        add_sea_slope(heights, s)

        print(f'Saving to {terrain_file_npy}...')
        np.save(terrain_file_npy, heights)

        # Store
        _heights = heights
        _heights.setflags(write=False)
        _not_sea = not_sea
        _not_sea.setflags(write=False)


def download_gagg(url, fname, print_to_screen=True):
    """
    Downloads the (160mb) ``gagg`` file containing the OS50 data from the
    Ordnance Survey.
    """
    if print_to_screen:
        print('Downloading terrain data...')
    url = urllib.request.urlopen(url)
    try:
        raw_data = url.read()
    finally:
        url.close()

    if print_to_screen:
        print('Writing...')
    with open(fname, 'wb') as f:
        f.write(raw_data)


def extract(basename, heights, resolution, print_to_screen=True):
    """
    Extracts data from an "ASCII Grid and GML (Grid)" file at ``basename``.zip,
    and extracts the data into numpy array ``heights``.

    The resolution of each file inside the zip must be the same, and must be
    specified as ``resolution``.
    """
    # Open zip file
    if print_to_screen:
        print(f'Reading from {terrain_file_zip}')
        i = 0

    t = nevis.Timer()
    with zipfile.ZipFile(terrain_file_zip, 'r') as f:

        # Find inner zip files
        zips = []
        for name in f.namelist():
            if os.path.splitext(name)[1] == '.zip':
                zips.append(name)
        zips.sort()

        # Read nested zip files
        for name in zips:
            read_nested_zip(f, name, heights, resolution)

            if print_to_screen:
                i += 1
                print('.', end=(None if i % 79 == 0 else ''))
                sys.stdout.flush()

    if print_to_screen:
        print(f'\nFinished, after {t.format()}')


def read_nested_zip(parent, name, heights, resolution):
    """
    Opens a zip-in-a-zip and reads and extracts any ``.asc`` files inside it.
    """
    # Open zip-in-a-zip
    with parent.open(name, 'r') as par:
        nested = io.BytesIO(par.read())
        with zipfile.ZipFile(nested) as f:

            # Scan for asc files
            for path in f.namelist():
                if os.path.splitext(path)[1] == '.asc':

                    # Read internal asc
                    with f.open(path, 'r') as asc:
                        try:
                            read_asc(asc, heights, resolution)
                        except Exception as e:
                            raise Exception(
                                f'Error reading {path} in {name}: {str(e)}.')


def read_asc(handle, heights, resolution):
    """
    Extracts data from a handle pointing to an already opened ``.asc`` file.
    """
    # Grab all text in one go and decode
    lines = handle.read().decode(ENC).splitlines()

    # Head lines
    def header(line, field):
        start = field.strip() + ' '
        n = len(start)
        if line[:n] != start:
            raise Exception(
                f'Unexpected header line. Got "{line}", expecting "{start}".')
        return int(line[n:])

    # Ncols and nrows
    ncols = header(lines[0], 'ncols')
    nrows = header(lines[1], 'nrows')

    # Offset
    xll = header(lines[2], 'xllcorner') // resolution
    yll = header(lines[3], 'yllcorner') // resolution

    # Resolution
    cellsize = header(lines[4], 'cellsize')
    if cellsize != resolution:
        raise Exception(
            f'Unexpected resolution. Got {cellsize}, expecting {resolution}.')

    # Optional missing value indicator
    missing = None
    offset = 5
    if lines[offset].startswith('nodata_value '):
        missing = lines[offset].split()[1:2]
        offset += 1

    # Read data
    data = np.genfromtxt(
        lines[offset:],
        delimiter=' ',
        missing_values=missing,
        filling_values=[np.nan],
        max_rows=nrows,
    )
    assert data.shape == (nrows, ncols)

    # Insert data into vector
    heights[yll:yll + nrows, xll:xll + ncols] = data[::-1, :]


def fix_sea_levels_in_odd_squares(heights):
    """
    Manually "lower" some areas just off the coast that have above sea-level
    heights in the data set.
    """
    # Fix sea level in NT68 (last checked on 2022-06-27)
    x, w = nevis.Coords.from_square_with_size('NT68')
    x, y = x.grid[0] // resolution, x.grid[1] // resolution
    w = w // resolution
    view = heights[y:y + w, x:x + w]
    view[view < 2.5] -= 2.1

    # Fix sea level in NR24, 34, 44 (last checked on 2022-06-27)
    x, w = nevis.Coords.from_square_with_size('NR24')
    x, y = x.grid[0] // resolution, x.grid[1] // resolution
    w = w // resolution
    view = heights[y:y + w, x:x + 3 * w]
    view[view < 0.2] -= 10

    # Fix sea level in NR33 (last checked on 2022-06-27)
    x, w = nevis.Coords.from_square_with_size('NR33')
    x, y = x.grid[0] // resolution, x.grid[1] // resolution
    w = w // resolution
    view = heights[y:y + w, x:x + w]
    view[view < 0.2] -= 10

    # Fix sea level in NR35 (last checked on 2022-06-27)
    x, w = nevis.Coords.from_square_with_size('NR35')
    x, y = x.grid[0] // resolution, x.grid[1] // resolution
    w = w // resolution
    view = heights[y:y + w, x:x + w]
    view[view < 0.2] -= 0.5

    # Fix sea level in NR56 (last checked on 2022-06-27)
    x, w = nevis.Coords.from_square_with_size('NR56')
    x, y = x.grid[0] // resolution, x.grid[1] // resolution
    w = w // resolution
    view = heights[y:y + w, x:x + w]
    view[view < 0.1] = -10

    # Fix sea level in NR57 (last checked on 2022-06-27)
    x, w = nevis.Coords.from_square_with_size('NR57')
    x, y = x.grid[0] // resolution, x.grid[1] // resolution
    w = w // resolution
    view = heights[y:y + w, x:x + w]
    view[view < 0.1] -= 0.5

    # Fix sea level in NR76 (last checked on 2022-06-27)
    x, w = nevis.Coords.from_square_with_size('NR76')
    x, y = x.grid[0] // resolution, x.grid[1] // resolution
    w = w // resolution
    view = heights[y:y + w, x:x + w]
    view[view < 0.1] -= 0.5

    # Fix sea level in NR65,75,64,74 (last checked on 2022-06-27)
    x, w = nevis.Coords.from_square_with_size('NR64')
    x, y = x.grid[0] // resolution, x.grid[1] // resolution
    w = w // resolution
    view = heights[y:y + 2 * w, x:x + 2 * w]
    view[view < 0.1] = -10
    del view


def save_cambridgeshire(heights):
    """
    Artificially raise the level of two river beds to stop the sea-level
    setting algorithm from treating e.g. Holme fen as "sea".
    """
    # Last checked 2023-08-27

    # Block river Great Ouse in TF 50, stopping a lot of flooding in
    # cambridgeshire.
    #heights[6047, 11183] = 0.01   # Worked 2022-06-27, but not 2023-08-31
    heights[6196, 11197] = 0.01

    # Block river Yare in TG 50, and Oulton Dyke in TM59, stopping lots of
    # flooding near Norwich.
    # Note: Both must be set to see an effect.
    heights[6151, 13041] = 0.01   # TG50
    heights[5851, 13013] = 0.01   # TM59


def inland_below_sea_level_points(heights, s, print_to_screen=True):
    """
    Return an array of (x, y) coordinates that are below sea-level, but not to
    be treated as sea.
    """
    y, x = np.nonzero(np.logical_and(heights < 0, heights > s))
    p = y * heights.shape[1] + x
    assert np.max(p) < 2**32, 'Not-sea point does not fit uint32'
    return np.array(p, dtype=np.uint32)


def set_sea_level(heights, s, print_to_screen=True):
    """
    Create a "mask" on the heights map, by setting all entries suspected to be
    sea to a fixed (very low) "sea level" value ``s``.
    """
    # Find a "sea mask", set all pixels to s
    if print_to_screen:
        print('Creating sea bitmask...')
    t = nevis.Timer()

    # Treat each square separately, starting bottom left and spiraling inwards
    # To iterate in this way, we first create a list of squares
    d = 80  # Trial and error shows this is near optimal
    squares = _spiral_squares(heights, d)

    # The edges of the map are all definitely sea, so set to s
    e = 1
    heights[:e, :] = s
    heights[-e:, :] = s
    heights[:, :e] = s
    heights[:, -e:] = s

    # Iterate over squares, and apply "sea mask" in each individually
    iters = 0
    ny, nx = heights.shape
    zmax = 100
    for z in range(zmax):

        skip = []
        changed = False
        for i, sq in enumerate(squares):
            if print_to_screen:
                iters += 1
                if iters % 100 == 0:
                    print('.', end=(None if (iters // 100) % 79 == 0 else ''))
                    sys.stdout.flush()

            # Create views of center of square, plus neighbours
            x0, y0 = sq
            x1, y1 = x0 + d, y0 + d

            # At the edge? Then translate by 1 pixel
            x0, y0 = max(x0, 1), max(y0, 1)
            x1, y1 = min(x1, nx - 1), min(y1, ny - 1)

            # Create view and views of neighbours
            v = heights[y0: y1, x0: x1]
            v0 = heights[y0 + 1: y1 + 1, x0: x1]  # Above
            v1 = heights[y0 - 1: y1 - 1, x0: x1]  # Below
            v2 = heights[y0: y1, x0 - 1: x1 - 1]  # Left
            v3 = heights[y0: y1, x0 + 1: x1 + 1]  # Right

            # Skip easy squares
            if np.all(v < 0):
                if (np.any(v == s) | np.any(v0 == s) | np.any(v1 == s)
                        | np.any(v2 == s) | np.any(v3 == s)):
                    v[:] = s
                    skip.append(i)
                    changed = True
                    # Don't skip if < 0 but no s neighbours _yet_
                continue
            elif np.all(v > 0):
                skip.append(i)
                continue

            kmax = 1000
            for k in range(1000):
                n = (v <= 0) & (v != s) & (
                    (v0 == s) | (v1 == s) | (v2 == s) | (v3 == s))
                if not np.any(n):
                    break
                v[n] = s
                changed = True
            if k + 1 == kmax:
                print('WARNING: Reached kmax')

        for i in reversed(skip):
            del squares[i]
        if not squares:
            break
        if not changed:
            break

    if z + 1 == zmax:
        print('WARNING: Reached zmax')

    if print_to_screen:
        print(f'\nFinished, after {t.format()}')


def add_sea_slope(heights, s, print_to_screen=True):
    """
    Make the sea slope downwards, as you move further away from the coast.

    The input should be a ``heights`` array in which all points known to be sea
    (and only those points) have been set to height ``s``.
    """
    print('Adding slope to sea bed')
    t = nevis.Timer()

    h = 0.01  # slope per square
    ny, nx = heights.shape

    # Create arrays of all neighbours above, below, left, and right of points
    v0 = heights[1:]
    v1 = heights[:-1]
    v2 = heights[:, 1:]
    v3 = heights[:, :-1]

    # Select all points next to a "sea" point, that are not themselves sea.
    yd, xd = np.nonzero(np.logical_and(v1 == s, v0 != s))
    yu, xu = np.nonzero(np.logical_and(v0 == s, v1 != s))
    yl, xl = np.nonzero(np.logical_and(v3 == s, v2 != s))
    yr, xr = np.nonzero(np.logical_and(v2 == s, v3 != s))
    y = np.concatenate((yd, yu + 1, yl, yr))
    x = np.concatenate((xd, xu, xl, xr + 1))
    # And set all these points heights to s (sea mask value) minus h
    heights[y, x] = s - h

    # Iteratively find all neighbours of the current selection that are still
    # at level s, or at a value lower than what this pass would set them to.
    # Then set these points to the current selection's height minus s.
    j = 0
    for i in range(ny * nx):
        div = 1 if len(x) > 300000 else (10 if len(x) > 50000 else 100)
        if i % div == 0:
            j += 1
            print('.', end=(None if j % 79 == 0 else ''))
            sys.stdout.flush()

        # Move down
        ok = np.nonzero(y > 0)
        yd, xd = y[ok] - 1, x[ok]
        ok = np.nonzero((heights[yd, xd] == s)
                        | (heights[yd, xd] < heights[yd + 1, xd] - h))
        yd, xd = yd[ok], xd[ok]
        heights[yd, xd] = heights[yd + 1, xd] - h

        # Move up
        ok = np.nonzero(y < ny - 1)
        yu, xu = y[ok] + 1, x[ok]
        ok = np.nonzero((heights[yu, xu] == s)
                        | (heights[yu, xu] < heights[yu - 1, xu] - h))
        yu, xu = yu[ok], xu[ok]
        heights[yu, xu] = heights[yu - 1, xu] - h

        # Move left
        ok = np.nonzero(x > 0)
        yl, xl = y[ok], x[ok] - 1
        ok = np.nonzero((heights[yl, xl] == s)
                        | (heights[yl, xl] < heights[yl, xl + 1] - h))
        yl, xl = yl[ok], xl[ok]
        heights[yl, xl] = heights[yl, xl + 1] - h

        # Move right
        ok = np.nonzero(x < nx - 1)
        yr, xr = y[ok], x[ok] + 1
        ok = np.nonzero((heights[yr, xr] == s)
                        | (heights[yr, xr] < heights[yr, xr - 1] - h))
        yr, xr = yr[ok], xr[ok]
        heights[yr, xr] = heights[yr, xr - 1] - h

        y = np.concatenate((yd, yu, yl, yr))
        x = np.concatenate((xd, xu, xl, xr))

        if len(y) == 0:
            break

    heights[heights < s] -= s

    print(f'\nFinished, after {t.format()}')


def _spiral_squares(heights, d):
    """
    Returns a list of squares (indicated by the lower-left corner) that cover
    the map, starting lower-left and spiraling inward. Each square has size
    ``d``.
    """
    squares = []
    x0, y0 = (0, 0)
    y1, x1 = heights.shape
    while x1 > x0:
        for x in range(x0, x1, d):
            squares.append((x, y0))
        y0 += d
        if y1 <= y0:
            break
        for y in range(y0, y1, d):
            squares.append((x1 - d, y))
        x1 -= d
        if x1 <= x0:
            break
        for x in range(x1 - d, x0 - d, -d):
            squares.append((x, y1 - d))
        y1 -= d
        if y1 <= y0:
            break
        for y in range(y1 - d, y0 - d, -d):
            squares.append((x0, y))
        x0 += d
    return squares



#
# SetupTools script for nevis
#
from setuptools import setup, find_packages


# Get version number
import os
import sys
sys.path.append(os.path.abspath('nevis'))
from _nevis_version import __version__ as version  # noqa
sys.path.pop()
del os, sys


# Load text for long description
with open('README.md', encoding='utf-8') as f:
    readme = f.read()


# Go!
setup(
    # See https://python-packaging.readthedocs.io/en/latest/index.html
    # for details of what goes in here.

    # Module name (lowercase)
    name='nevis',

    # Version
    version=version,

    # Description
    description=('Presents the landscape of Great Britain as a testbed for'
                 ' optimisation and sampling methods.'),
    long_description=readme,
    long_description_content_type='text/markdown',

    # Author and license
    license='BSD 3-clause license',
    author='The Where Is Ben Nevis team',
    author_email='michael.clerx@nottingham.ac.uk',

    # Project URLs (only first is required, rest is for PyPI)
    url='https://github.com/CardiacModelling/BenNevis',
    project_urls={
        'Bug Tracker': 'https://github.com/CardiacModelling/BenNevis/issues/',
        'Documentation': 'https://nevis.readthedocs.io/',
        'Source Code': 'https://github.com/CardiacModelling/BenNevis/',
    },

    # Classifiers for pypi
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Mathematics',
    ],

    # Packages to include
    packages=find_packages(include=('nevis', 'nevis.*')),

    # Include non-python files (via MANIFEST.in)
    include_package_data=True,

    # List of dependencies
    install_requires=[
        'bnglonlat',
        'matplotlib>=1.5',
        'numpy',
        'pykml',
        'scipy',
        'setuptools',
        'zipfile-deflate64',
    ],

    # Optional extras
    extras_require={
        'dev': [
            #'coverage',                 # Coverage checking
            'flake8>=3',                # Style checking
            'sphinx>=1.5, !=1.7.3',     # Doc generation
        ],
        'extras': [
            'convertbng',       # Accurate version of bnglonlat
            'pillow',           # To check generated image sizes (PIL)
        ],
    },

    # Python version
    python_requires='>=3.6',
)



#!/usr/bin/env python3
"""
Provides utility methods (i.e. methods not directly related to ``nevis``).
"""
import timeit

import nevis


class Timer(object):
    """
    Provides accurate timing.

    Example
    -------
    ::

        timer = nevis.Timer()
        print(timer.format(timer.time()))

    """

    def __init__(self, output=None):
        self._start = timeit.default_timer()
        self._methods = {}

    def format(self, time=None):
        """
        Formats a (non-integer) number of seconds, returns a string like
        "5 weeks, 3 days, 1 hour, 4 minutes, 9 seconds", or "0.0019 seconds".
        """
        if time is None:
            time = self.time()
        if time < 1e-2:
            return str(time) + ' seconds'
        elif time < 60:
            return str(round(time, 2)) + ' seconds'
        output = []
        time = int(round(time))
        units = [
            (604800, 'week'),
            (86400, 'day'),
            (3600, 'hour'),
            (60, 'minute'),
        ]
        for k, name in units:
            f = time // k
            if f > 0 or output:
                output.append(str(f) + ' ' + (name if f == 1 else name + 's'))
            time -= f * k
        output.append('1 second' if time == 1 else str(time) + ' seconds')
        return ', '.join(output)

    def reset(self):
        """
        Resets this timer's start time.
        """
        self._start = timeit.default_timer()

    def time(self):
        """
        Returns the time (in seconds) since this timer was created, or since
        meth:`reset()` was last called.
        """
        return timeit.default_timer() - self._start


#
# Version-related methods
#
def howdy(name='version'):
    """ Say hi the old fashioned way. """
    print(r'')
    print(r'                |>          ')
    print(r' Starting Ben   |   Nevis   ')
    print(r'               / \    ' + name)
    print(r'            /\/---\     ' + nevis.__version__)
    print(r'           /---    \/\      ')
    print(r'        /\/   /\   /  \     ')
    print(r'     /\/  \  /  \_/    \    ')
    print(r'    /      \/           \   ')


def print_result(x, y, z):
    """
    Print information about an optimisation result.

    Arguments:

    ``x``
        The x-coordinate of the result.
    ``y``
        The y-coordinate of the result.
    ``z``
        The z-coordinate of the result.
    """

    coords = nevis.Coords(gridx=x, gridy=y)
    hill, distance = nevis.Hill.nearest(coords)
    print('Congratulations!' if distance < 100 else (
        'Good job!' if distance < 1000 else 'Interesting!'))
    print(f'You landed at an altitude of {round(z)}m.')
    print(f'  {coords.opentopomap}')
    dm = (
        f'{round(distance)}m' if distance < 1000
        else f'{round(distance / 1000, 1)}km'
    )
    print(f'You are {dm} from the nearest named hill top, "{hill.name}",')
    print(f'  ranked the {hill.ranked} heighest in GB.')
    photo = hill.photo()
    if photo:
        print('  ' + photo)


def generate_kml(path, labels=None, trajectory=None, points=None,):
    """
    Generate a KML (Keyhole Markup Language, used to display geographic data
    in an Earth browser such as Google Earth) file from given data.

    For a description of the KML format, see
    https://developers.google.com/kml/documentation/kml_tut

    Arguments:

    ``path``
        The path of the file to write to.
    ``labels``
        An optional dictionary mapping string labels to points (tuples in
        meters or Coords objects) that will be marked on the map. The points
        will be shown as green pinpoints with text labels beside them.
    ``trajectory``
        An optional array of shape ``(n_points, 2)`` indicating the trajectory
        to be plotted (points specified in meters). All the points along the
        trajectory will be shown as small orange pinpoints with labels of
        their index prefixed with "T" (e.g. "T3"). The trajectory itself will
        be shown as a red curve extended down to the ground connecting
        adjacent points.
    ``points``
        An optional array of shape ``(n_points, 2)`` indicating points on the
        map (points specified in meters). All the points will be shown as small
        blue pinpoints with labels of their index prefixed with "P" (e.g.
        "P3").

    ``lables``, ``trajectory``, and ``points`` can be used simultaneously.
    """
    from pykml.factory import KML_ElementMaker as KML
    from lxml import etree

    def make_coords(x, y):
        """
        Convert a pair of coordinates (in meters) to a KML coordinates string.
        """
        lat, lon = nevis.Coords(x, y).latlong
        return f'{lon},{lat}'

    marks = []
    if labels is not None:
        for label, p in labels.items():
            if isinstance(p, nevis.Coords):
                p = p.grid
            x, y = p

            marks.append(
                KML.Placemark(
                    KML.name(label),
                    KML.Point(
                        KML.coordinates(make_coords(x, y))
                    ),
                    KML.styleUrl('#label_icon_style'),
                )
            )

    if points is not None:
        for i, (x, y) in enumerate(points):
            marks.append(
                KML.Placemark(
                    KML.name(f'P{i}'),
                    KML.Point(
                        KML.coordinates(make_coords(x, y))
                    ),
                    KML.styleUrl('#points_icon_style'),
                )
            )

    if trajectory is not None:
        for i, (x, y) in enumerate(trajectory):
            marks.append(
                KML.Placemark(
                    KML.name(f'T{i}'),
                    KML.Point(
                        KML.coordinates(make_coords(x, y))
                    ),
                    KML.styleUrl('#trajectory_icon_style'),
                )
            )

        marks.append(
            KML.Placemark(
                KML.name('Trajectory'),
                KML.styleUrl('#line_style'),
                KML.LineString(
                    KML.coordinates(
                        ' '.join([make_coords(x, y) for x, y in trajectory])
                    ),
                    KML.extrude(1),
                    KML.tessellate(1),
                )
            )
        )

    icon_url = 'http://maps.google.com/mapfiles/kml/paddle/wht-blank.png'

    doc = KML.kml(
        KML.Document(
            KML.name('Nevis Project'),
            KML.Style(
                KML.LineStyle(
                    KML.color('ff1400FF'),
                    KML.width(3),
                ),
                id="line_style"
            ),
            KML.Style(
                KML.IconStyle(
                    KML.Icon(
                        KML.href(icon_url),
                    ),
                    KML.color('ff00FF14'),
                ),
                id="label_icon_style"
            ),
            KML.Style(
                KML.IconStyle(
                    KML.Icon(
                        KML.href(icon_url),
                    ),
                    KML.scale(0.5),
                    KML.color('ffFF7800'),
                ),
                KML.LabelStyle(
                    KML.scale(0.5),
                ),
                id="points_icon_style"
            ),
            KML.Style(
                KML.IconStyle(
                    KML.Icon(
                        KML.href(icon_url),
                    ),
                    KML.scale(0.5),
                    KML.color('ff1e90ff'),
                ),
                KML.LabelStyle(
                    KML.scale(0.5),
                ),
                id="trajectory_icon_style"
            ),
            *marks,
        )
    )

    with open(path, 'w') as fobj:
        fobj.write(
            etree.tostring(doc, pretty_print=True).decode('utf-8'))
